#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
import re
from fractions import Fraction

from ase.geometry import crystal_structure_from_cell
from ase.dft.kpoints import special_paths, parse_path_string, get_special_points
try:
    from ase.dft.kpoints import special_points
except ImportError:
    from ase.dft.kpoints import sc_special_points as special_points

from phonopy.phonon.band_structure import BandStructure
from phonopy.units import EvTokJmol, VaspToTHz


def generate_kpath_parameters(points, paths, npoints):
    k_points = []
    for p in paths:
        k_points.append([points[k] for k in p])
        for index in range(len(p)):
            if p[index] == 'G':
                p[index] = 'Γ'
    parameters = []
    for h, seg in enumerate(k_points):
        for i, path in enumerate(seg):
            parameter = {}
            parameter['npoints'] = npoints
            parameter['startname'] = paths[h][i]
            if i == 0 and len(seg) > 2:
                parameter['kstart'] = path
                parameter['kend'] = seg[i + 1]
                parameter['endname'] = paths[h][i + 1]
                parameters.append(parameter)
            elif i == (len(seg) - 2):
                parameter['kstart'] = path
                parameter['kend'] = seg[i + 1]
                parameter['endname'] = paths[h][i + 1]
                parameters.append(parameter)
                break
            else:
                parameter['kstart'] = path
                parameter['kend'] = seg[i + 1]
                parameter['endname'] = paths[h][i + 1]
                parameters.append(parameter)
    return parameters


def read_kpath(filename):
    with open(filename) as f:
        string = f.read()

        labels = re.search(r'BAND_LABELS\s*=\s*(.+)', string)
        try:
            labels = labels.group(1).strip().split()
        except Exception:
            return

        points = re.search(r'BAND\s*=\s*(.+)', string)
        try:
            points = points.group(1)
            points = [float(Fraction(p)) for p in points.split()]
            points = np.reshape(points, (len(labels), 3))
            points = {labels[i]: points[i] for i in range(len(labels))}
        except Exception:
            return

        npoints = re.search(r'BAND_POINTS\s*\=\s*(\d+)', string)
        if npoints is not None:
            npoints = int(npoints.group(1))
        else:
            npoints = 100

    return generate_kpath_parameters(points, [labels], npoints)


def generate_kpath_ase(cell, symprec):

    eig_val_max = np.real(np.linalg.eigvals(cell)).max()
    eps = eig_val_max * symprec

    lattice = crystal_structure_from_cell(cell, eps)
    paths = special_paths.get(lattice, None)
    if paths is None:
        paths = special_paths['orthorhombic']
    paths = parse_path_string(special_paths[lattice])
    points = special_points.get(lattice)
    if points is None:
        try:
            points = get_special_points(cell)
        except Exception:
            return []

    return generate_kpath_parameters(points, paths, 100)


class PhononProperties():
    def __init__(self, phonopy_obj, logger, **kwargs):
        self.logger = logger
        self.phonopy_obj = phonopy_obj
        self.t_max = kwargs.get('t_max', 1000)
        self.t_min = kwargs.get('t_min', 0)
        self.t_step = kwargs.get('t_step', 100)
        self.band_conf = kwargs.get('band_conf')

        self.n_atoms = len(phonopy_obj.unitcell)

        k_mesh = kwargs.get('k_mesh', 30)
        mesh_density = (2 * k_mesh ** 3) / self.n_atoms
        mesh_number = np.round(mesh_density**(1. / 3.))
        self.mesh = [mesh_number, mesh_number, mesh_number]

        self.n_atoms_supercell = len(phonopy_obj.supercell)

    def get_bandstructure(self):
        phonopy_obj = self.phonopy_obj

        frequency_unit_factor = VaspToTHz
        is_eigenvectors = False

        unit_cell = phonopy_obj.unitcell.get_cell()
        sym_tol = phonopy_obj.symmetry.tolerance
        if self.band_conf is not None:
            parameters = read_kpath(self.band_conf)
        else:
            parameters = generate_kpath_ase(unit_cell, sym_tol)
        if not parameters:
            return None, None, None

        # Distances calculated in phonopy.band_structure.BandStructure object
        # are based on absolute positions of q-points in reciprocal space
        # as calculated by using the cell which is handed over during instantiation.
        # Fooling that object by handing over a "unit cell" diag(1,1,1) instead clashes
        # with calculation of non-analytical terms.
        # Hence generate appropriate distances and special k-points list based on fractional
        # coordinates in reciprocal space (to keep backwards compatibility with previous
        # FHI-aims phonon implementation).
        bands = []
        bands_distances = []
        distance = 0.0
        bands_special_points = [distance]
        bands_labels = []
        label = parameters[0]["startname"]
        for b in parameters:
            kstart = np.array(b["kstart"])
            kend = np.array(b["kend"])
            npoints = b["npoints"]
            dk = (kend - kstart) / (npoints - 1)
            bands.append([(kstart + dk * n) for n in range(npoints)])
            dk_length = np.linalg.norm(dk)

            for n in range(npoints):
                bands_distances.append(distance + dk_length * n)

            distance += dk_length * (npoints - 1)
            bands_special_points.append(distance)
            label = [b["startname"], b["endname"]]
            bands_labels.append(label)

        bs_obj = BandStructure(
            bands, phonopy_obj.dynamical_matrix, with_eigenvectors=is_eigenvectors,
            factor=frequency_unit_factor)

        freqs = bs_obj.get_frequencies()

        return np.array(freqs), np.array(bands), np.array(bands_labels)

    def get_dos(self):
        phonopy_obj = self.phonopy_obj
        mesh = self.mesh

        phonopy_obj.set_mesh(mesh, is_gamma_center=True)
        q_points = phonopy_obj.get_mesh()[0]
        phonopy_obj.set_qpoints_phonon(q_points, is_eigenvectors=False)

        frequencies = phonopy_obj.get_qpoints_phonon()[0]
        self.frequencies = np.array(frequencies)
        min_freq = min(np.ravel(frequencies))
        max_freq = max(np.ravel(frequencies)) + max(np.ravel(frequencies)) * 0.05

        phonopy_obj.set_total_DOS(
            freq_min=min_freq, freq_max=max_freq, tetrahedron_method=True)
        f, dos = phonopy_obj.get_total_DOS()

        return f, dos

    def get_thermodynamical_properties(self):
        phonopy_obj = self.phonopy_obj

        phonopy_obj.set_mesh(self.mesh, is_gamma_center=True)
        phonopy_obj.set_thermal_properties(
            t_step=self.t_step, t_max=self.t_max, t_min=self.t_min)
        T, fe, entropy, cv = phonopy_obj.get_thermal_properties()
        kJmolToEv = 1.0 / EvTokJmol
        fe = fe * kJmolToEv
        JmolToEv = kJmolToEv / 1000
        cv = JmolToEv * cv
        return T, fe, entropy, cv



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from elasticsearch_dsl import Text, Keyword

from nomad.metainfo import MSection, Section, SubSection, Quantity, MEnum, Reference, Package
from nomad.metainfo.search_extension import Search

# This is usally defined automatically when the first metainfo definition is evaluated, but
# due to the next imports requireing the m_package already, this would be too late.
m_package = Package()

from .metainfo.common_dft import section_k_band, section_dos, section_thermodynamical_properties, FastAccess  # noqa


class WyckoffVariables(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        description="""
        Contains the variables associated with a Wyckoff set.
        """
    )
    x = Quantity(
        type=np.dtype(np.float64),
        description="""
        The x variable if present.
        """
    )
    y = Quantity(
        type=np.dtype(np.float64),
        description="""
        The y variable if present.
        """
    )
    z = Quantity(
        type=np.dtype(np.float64),
        description="""
        The z variable if present.
        """
    )


class WyckoffSet(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        description="""
        Section for storing Wyckoff set information.
        """
    )
    wyckoff_letter = Quantity(
        type=str,
        description="""
        The Wyckoff letter for this set.
        """
    )
    indices = Quantity(
        type=np.dtype('i4'),
        shape=["1..*"],
        description="""
        Indices of the atoms belonging to this group.
        """
    )
    element = Quantity(
        type=str,
        description="""
        Chemical element at this Wyckoff position.
        """
    )
    variables = SubSection(sub_section=WyckoffVariables.m_def, repeats=False, categories=[FastAccess])


class LatticeParameters(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        description="""
        Lattice parameters of the idealized cell. The lattice parameters can
        only be reported consistently after idealization and may not perfectly
        correspond to the original simulation cell.
        """,
        a_search="lattice_parameters"
    )
    a = Quantity(
        type=float,
        description="""
        Length of the first basis vector.
        """,
        a_search=Search()
    )
    b = Quantity(
        type=float,
        description="""
        Length of the second basis vector.
        """,
        a_search=Search()
    )
    c = Quantity(
        type=float,
        description="""
        Length of the third basis vector.
        """,
        a_search=Search()
    )
    alpha = Quantity(
        type=float,
        description="""
        Angle between second and third basis vector.
        """,
        a_search=Search()
    )
    beta = Quantity(
        type=float,
        description="""
        Angle between first and third basis vector.
        """,
        a_search=Search()
    )
    gamma = Quantity(
        type=float,
        description="""
        Angle between first and second basis vector.
        """,
        a_search=Search()
    )


class IdealizedStructure(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        description="""
        Contains structural information for an idealized representation of the
        material used in the calculation. This idealization is used for
        visualizing the material and for calculating the structural properties.
        The properties of the idealized structure may slightly vary from the
        original structure used in the calculation.
        """,
        a_search="idealized_structure",
    )
    atom_labels = Quantity(
        type=str,
        shape=['1..*'],
        description="""
        Type (element, species) of each atom.
        """
    )
    atom_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description="""
        Atom positions given in coordinates that are relative to the idealized
        cell.
        """
    )
    lattice_vectors = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description="""
        Lattice vectors of the idealized structure. For bulk materials it is
        the Bravais cell. This cell is representative and is idealized to match
        the detected symmetry properties.
        """
    )
    lattice_vectors_primitive = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        description="""
        Lattice vectors of the the primitive unit cell in a form to be visualized
        within the idealized cell. This cell is representative and is
        idealized to match the detected symmemtry properties.
        """
    )
    periodicity = Quantity(
        type=np.bool_,
        shape=[3],
        description="""
        Automatically detected true periodicity of each lattice direction. May
        not correspond to the periodicity used in the calculation.
        """
    )
    number_of_atoms = Quantity(
        type=int,
        description="""
        Number of atoms in the idealized structure."
        """
    )
    cell_volume = Quantity(
        type=np.dtype(np.float64),
        unit="m ** 3",
        description="""
        Volume of the idealized cell. The cell volume can only be reported
        consistently after idealization and may not perfectly correspond to the
        original simulation cell.
        """,
        a_search=Search()
    )
    wyckoff_sets = SubSection(sub_section=WyckoffSet.m_def, repeats=True, categories=[FastAccess])
    lattice_parameters = SubSection(sub_section=LatticeParameters.m_def, categories=[FastAccess])


class Bulk(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        a_search="bulk",
        description="""
        Contains information that is specific to bulk crystalline materials.
        """
    )
    bravais_lattice = Quantity(
        type=MEnum("aP", "mP", "mS", "oP", "oS", "oI", "oF", "tP", "tI", "hR", "hP", "cP", "cI", "cF"),
        description="""
        The Bravais lattice type in the Pearson notation, where the first
        lowercase letter indicates the crystal system, and the second uppercase
        letter indicates the lattice type. The value can only be one of the 14
        different Bravais lattices in three dimensions.

        Crystal system letters:

        a = Triclinic
        m = Monoclinic
        o = Orthorhombic
        t = Tetragonal
        h = Hexagonal and Trigonal
        c = Cubic

        Lattice type letters:

        P = Primitive
        S (A, B, C) = One side/face centred
        I = Body centered
        R = Rhombohedral centring
        F = All faces centred
        """,
        a_search=Search()
    )
    crystal_system = Quantity(
        type=MEnum("triclinic", "monoclinic", "orthorhombic", "tetragonal", "trigonal", "hexagonal", "cubic"),
        description="""
        The detected crystal system. One of seven possibilities in three dimensions.
        """,
        a_search=Search()
    )
    has_free_wyckoff_parameters = Quantity(
        type=bool,
        description="""
        Whether the material has any Wyckoff sites with free parameters. If a
        materials has free Wyckoff parameters, at least some of the atoms are
        not bound to a particular location in the structure but are allowed to
        move with possible restrictions set by the symmetry.
        """,
        a_search=Search()
    )
    point_group = Quantity(
        type=MEnum("1", "-1", "2", "m", "2/m", "222", "mm2", "mmm", "4", "-4", "4/m", "422", "4mm", "-42m", "4/mmm", "3", "-3", "32", "3m", "-3m", "6", "-6", "6/m", "622", "6mm", "-6m2", "6/mmm", "23", "m-3", "432", "-43m", "m-3m"),
        description="""
        Point group in Hermann-Mauguin notation, part of crystal structure
        classification. There are 32 point groups in three dimensional space.
        """,
        a_search=Search()
    )
    space_group_number = Quantity(
        type=int,
        description="""
        Integer representation of the space group, part of crystal structure
        classification, part of material definition.
        """,
        a_search=Search()
    )
    space_group_international_short_symbol = Quantity(
        type=str,
        description="""
        International short symbol notation of the space group.
        """,
        a_search=Search()
    )
    structure_prototype = Quantity(
        type=str,
        description="""
        The prototypical material for this crystal structure.
        """,
        a_search=Search()
    )
    structure_type = Quantity(
        type=str,
        description="""
        Classification according to known structure type, considering the point
        group of the crystal and the occupations with different atom types.
        """,
        a_search=Search()
    )
    strukturbericht_designation = Quantity(
        type=str,
        description="""
        Classification of the material according to the historically grown "strukturbericht".
        """,
        a_search=Search()
    )


class Material(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        a_search="material",
        description="""
        Contains an overview of the type of material that was detected in this
        entry.
        """
    )
    material_type = Quantity(
        type=MEnum(bulk="bulk", two_d="2D", one_d="1D", unavailable="unavailable"),
        description="""
        "Broad structural classification for the material, e.g. bulk, 2D, 1D... ",
        """,
        a_search=Search()
    )
    material_id = Quantity(
        type=str,
        description="""
        A fixed length, unique material identifier in the form of a hash
        digest.
        """,
        a_search=Search(
            group='materials_grouped',
            metric='cardinality', metric_name='materials',
            description='Search for a particular material by its id.')
    )
    material_name = Quantity(
        type=str,
        description="""
        Most meaningful name for a material if one could be assigned
        """,
        a_search=Search()
    )
    material_classification = Quantity(
        type=str,
        description="""
        Contains the compound class and classification of the material
        according to springer materials in JSON format.
        """
    )
    formula = Quantity(
        type=str,
        description="""
        Formula giving the composition and occurrences of the elements in the
        Hill notation. For periodic materials the formula is calculated fom the
        primitive unit cell.
        """,
        a_search=Search()
    )
    formula_reduced = Quantity(
        type=str,
        description="""
        Formula giving the composition and occurrences of the elements in the
        Hill notation where the number of occurences have been divided by the
        greatest common divisor.
        """,
        a_search=Search()
    )
    species_and_counts = Quantity(
        type=str,
        description="""
        The formula separated into individual terms containing both the atom
        type and count. Used for searching parts of a formula.
        """,
        a_search=Search(mapping=Text(multi=True, fields={'keyword': Keyword()}))
    )
    species = Quantity(
        type=str,
        description="""
        The formula separated into individual terms containing only unique atom
        species. Used for searching materials containing specific elements.
        """,
        a_search=Search(mapping=Text(multi=True, fields={'keyword': Keyword()}))
    )

    # Bulk-specific properties
    bulk = SubSection(sub_section=Bulk.m_def, repeats=False, categories=[FastAccess])

    # The idealized structure for this material
    idealized_structure = SubSection(sub_section=IdealizedStructure.m_def, repeats=False, categories=[FastAccess])


class Method(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        a_search="method",
        description="""
        Contains an overview of the methodology that was detected in this
        entry.
        """
    )
    method_type = Quantity(
        type=MEnum("DFT", "GW", "unavailable", DFTU="DFT+U"),
        description="""
        Generic name for the used methodology.
        """,
        a_search=Search()
    )
    core_electron_treatment = Quantity(
        type=MEnum("full all electron", "all electron frozen core", "pseudopotential", "unavailable"),
        description="""
        How the core electrons are described.
        """,
        a_search=Search()
    )
    functional_long_name = Quantity(
        type=str,
        description="""
        Full identified for the used exchange-correlation functional.
        """,
        a_search=Search()
    )
    functional_type = Quantity(
        type=str,
        description="""
        Basic type of the used exchange-correlation functional.
        """,
        a_search=Search()
    )
    method_id = Quantity(
        type=str,
        description="""
        A fixed length, unique method identifier in the form of a hash digest.
        The hash is created by using several method settings as seed. This hash
        is only defined if a set of well-defined method settings is available
        for the used program.
        """
    )
    group_eos_id = Quantity(
        type=str,
        description="""
        A fixed length, unique identifier for equation-of-state calculations.
        Only calculations within the same upload and with a method hash
        available will be grouped under the same hash.
        """,
        a_search=Search()
    )
    group_parametervariation_id = Quantity(
        type=str,
        description="""
        A fixed length, unique identifier for calculations where structure is
        identical but the used computational parameters are varied. Only
        calculations within the same upload and with a method hash available
        will be grouped under the same hash.
        """,
        a_search=Search()
    )
    gw_starting_point = Quantity(
        type=str,
        description="""
        The exchange-correlation functional that was used as a starting point
        for this GW calculation.
        """
    )
    gw_type = Quantity(
        type=MEnum("G0W0", "scGW"),
        description="""
        Basic type of GW calculation.
        """
    )
    smearing_kind = Quantity(
        type=str,
        description="""
        Smearing function used for the electronic structure calculation.
        """
    )
    smearing_parameter = Quantity(
        type=np.dtype(np.float64),
        description="""
        Parameter for smearing, usually the width.
        """
    )


class Calculation(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        a_search="calculation",
        description="""
        Contains an overview of the type of calculation that was detected in
        this entry.
        """
    )
    calculation_type = Quantity(
        type=MEnum(
            single_point="single point",
            geometry_optimization="geometry optimization",
            molecular_dynamics="molecular dynamics",
            phonon_calculation="phonon calculation",
            elastic_constants="elastic constants",
            qha_calculation="QHA calculation",
            gw_calculation="GW calculation",
            equation_of_state="equation of state",
            parameter_variation="parameter variation",
            unavailable="unavailable"),
        description="""
        Defines the type of calculation that was detected for this entry.
        """,
        a_search=Search()
    )


class Energies(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        a_search="energies",
        description="""
        Contains different types of energies extracted from this entry. The
        energies are extracted from a representative calculation: for geometry
        optimization it is the last optimization step.
        """
    )
    energy_total = Quantity(
        type=np.dtype(np.float64),
        unit="eV",
        description="""
        Total energy.
        """,
        a_search=Search()
    )
    energy_total_T0 = Quantity(
        type=np.dtype(np.float64),
        unit="eV",
        description="""
        Total energy projected to T=0.
        """,
        a_search=Search()
    )
    energy_free = Quantity(
        type=np.dtype(np.float64),
        unit="eV",
        description="""
        Free energy.
        """,
        a_search=Search()
    )


class Properties(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        a_search="properties",
        description="""
        Contains derived physical properties that are specific to the NOMAD
        Encyclopedia.
        """
    )
    atomic_density = Quantity(
        type=np.dtype(np.float64),
        unit="1 / m ** 3",
        description="""
        Atomic density of the material (atoms/volume)."
        """,
        a_search=Search()
    )
    mass_density = Quantity(
        type=np.dtype(np.float64),
        unit="kg / m ** 3",
        description="""
        Mass density of the material.
        """,
        a_search=Search()
    )
    band_gap = Quantity(
        type=np.dtype(np.float64),
        unit="eV",
        description="""
        Band gap value. If multiple spin channels are present, this value is
        taken from the channel with smallest band gap value.
        """,
        a_search=Search()
    )
    band_gap_direct = Quantity(
        type=bool,
        description="""
        Whether band gap is direct or not. If multiple spin channels are
        present, this value is taken from the channel with smallest band gap
        value.
        """,
        a_search=Search()
    )
    energies = SubSection(sub_section=Energies.m_def, repeats=False, categories=[FastAccess], a_search=Search())
    electronic_band_structure = Quantity(
        type=Reference(section_k_band.m_def),
        shape=[],
        description="""
        Reference to an electronic band structure.
        """,
        a_search=Search(value=lambda section: section.electronic_band_structure.m_proxy_value if section.electronic_band_structure is not None else None, mapping=Keyword())
    )
    electronic_dos = Quantity(
        type=Reference(section_dos.m_def),
        shape=[],
        description="""
        Reference to an electronic density of states.
        """,
        a_search=Search(value=lambda section: section.electronic_dos.m_proxy_value if section.electronic_dos is not None else None, mapping=Keyword())
    )
    phonon_band_structure = Quantity(
        type=Reference(section_k_band.m_def),
        shape=[],
        description="""
        Reference to a phonon band structure.
        """,
        a_search=Search(value=lambda section: section.phonon_band_structure.m_proxy_value if section.phonon_band_structure is not None else None, mapping=Keyword())
    )
    phonon_dos = Quantity(
        type=Reference(section_dos.m_def),
        shape=[],
        description="""
        Reference to a phonon density of states.
        """,
        a_search=Search(value=lambda section: section.phonon_dos.m_proxy_value if section.phonon_dos is not None else None, mapping=Keyword())
    )
    thermodynamical_properties = Quantity(
        type=Reference(section_thermodynamical_properties.m_def),
        shape=[],
        description="""
        Reference to a section containing thermodynamical properties.
        """,
        a_search=Search(value=lambda section: section.thermodynamical_properties.m_proxy_value if section.thermodynamical_properties is not None else None, mapping=Keyword())
    )


class EncyclopediaMetadata(MSection):
    m_def = Section(
        a_flask=dict(skip_none=True),
        a_search=Search(),
        description="""
        Section which stores information for the NOMAD Encyclopedia.
        """
    )
    material = SubSection(sub_section=Material.m_def, repeats=False, categories=[FastAccess], a_search=Search())
    method = SubSection(sub_section=Method.m_def, repeats=False, categories=[FastAccess], a_search=Search())
    properties = SubSection(sub_section=Properties.m_def, repeats=False, categories=[FastAccess], a_search=Search())
    calculation = SubSection(sub_section=Calculation.m_def, repeats=False, categories=[FastAccess], a_search=Search())
    status = Quantity(
        type=MEnum("success", "unsupported_material_type", "unsupported_method_type", "unsupported_calculation_type", "invalid_metainfo", "failure"),
        description="""
        The final Encyclopedia processing status for this entry. The meaning of the status is as follows:

        | Status                           | Description                                                                   |
        | -------------------------------- | ----------------------------------------------------------------------------- |
        | `"success"`                      | Processed successfully                                                        |
        | `"unsupported_material_type"`    | The detected material type is currently not supported by the Encyclopedia.    |
        | `"unsupported_calculation_type"` | The detected calculation type is currently not supported by the Encyclopedia. |
        | `"invalid_metainfo"`             | The entry could not be processed due to missing or invalid metainfo.          |
        | `"failure"`                      | The entry could not be processed due to an unexpected exception.              |
        """,
        a_search=Search()
    )



from typing import Optional, Tuple, List, Union, Dict, Set, Any
from fastapi import HTTPException
from pydantic import create_model
from elasticsearch_dsl import Search, Q
from datetime import datetime
import numpy as np

from optimade.filterparser import LarkParser
from optimade.server.entry_collections import EntryCollection
from optimade.server.query_params import EntryListingQueryParams, SingleEntryQueryParams
from optimade.server.exceptions import BadRequest
from optimade.server.mappers import StructureMapper
from optimade.models import StructureResource, StructureResourceAttributes
from optimade.models.utils import OptimadeField, SupportLevel
from optimade.server.schemas import ENTRY_INFO_SCHEMAS

from nomad import datamodel, files, search, utils, metainfo, config
from nomad.normalizing.optimade import (
    optimade_chemical_formula_reduced, optimade_chemical_formula_anonymous,
    optimade_chemical_formula_hill)

from .filterparser import _get_transformer as get_transformer
from .common import provider_specific_fields


logger = utils.get_logger(__name__)
float64 = np.dtype('float64')


class StructureResourceAttributesByAlias(StructureResourceAttributes):
    nmd_entry_page_url: Optional[str] = OptimadeField(
        None,
        alias='_nmd_entry_page_url',
        description='The url for the NOMAD gui entry page for this structure.',
        support=SupportLevel.OPTIONAL)

    nmd_raw_file_download_url: Optional[str] = OptimadeField(
        None,
        alias='_nmd_raw_file_download_url',
        description='The url to download all calculation raw files as .zip file.',
        support=SupportLevel.OPTIONAL)

    nmd_archive_url: Optional[str] = OptimadeField(
        None,
        alias='_nmd_archive_url',
        description='The url to the NOMAD archive json of this structure.',
        support=SupportLevel.OPTIONAL)

    def dict(self, *args, **kwargs):
        kwargs['by_alias'] = True
        return super().dict(*args, **kwargs)


def create_nomad_structure_resource_attributes_cls():
    fields: Dict[str, Tuple[type, OptimadeField]] = {}

    for name, search_quantity in provider_specific_fields():
        quantity = search_quantity.definition

        pydantic_type: type
        if not quantity.is_scalar:
            pydantic_type = list
        elif quantity.type in [str, int, float, bool]:
            pydantic_type = quantity.type if quantity.type != float64 else float
        elif quantity.type == metainfo.Datetime:
            pydantic_type = datetime
        elif isinstance(quantity.type, metainfo.MEnum):
            pydantic_type = str
        elif isinstance(quantity.type, metainfo.Reference):
            continue
        else:
            raise NotImplementedError('Search quantity type not support in optimade API')

        field = Optional[pydantic_type], OptimadeField(
            None,
            alias=f'_nmd_{name}',
            sortable=False,
            description=quantity.description if quantity.description else 'Not available. Will be added soon.',
            support=SupportLevel.OPTIONAL,
            queryable=SupportLevel.OPTIONAL)

        fields[f'nmd_{name}'] = field

    return create_model(
        'NomadStructureResourceAttributes',
        __base__=StructureResourceAttributesByAlias,
        **fields)


NomadStructureResourceAttributes = create_nomad_structure_resource_attributes_cls()


class NomadStructureResource(StructureResource):
    attributes: NomadStructureResourceAttributes  # type: ignore


ENTRY_INFO_SCHEMAS['structures'] = NomadStructureResource.schema


class StructureCollection(EntryCollection):

    def __init__(self):
        super().__init__(
            resource_cls=NomadStructureResource,
            resource_mapper=StructureMapper,
            transformer=get_transformer(nomad_properties='dft', without_prefix=False))

        self.parser = LarkParser(version=(1, 0, 0), variant="default")

        # check aliases do not clash with mongo operators
        self._check_aliases(self.resource_mapper.all_aliases())
        self._check_aliases(self.resource_mapper.all_length_aliases())

        self.client = Search

    def _base_search_request(self):
        request = search.SearchRequest().owner('public', None)
        request.search_parameter('processed', True)
        # TODO use the elastic annotations when done
        request.query(Q('exists', field='dft.optimade.elements'))
        return request

    def __len__(self) -> int:
        # TODO cache
        return self._base_search_request().execute()['total']

    def count(self, **kwargs) -> int:
        # This seams solely mongodb specific
        raise NotImplementedError()

    def find(
            self,
            params: Union[EntryListingQueryParams, SingleEntryQueryParams]) \
            -> Tuple[List[StructureResource], int, bool, set]:

        criteria = self.handle_query_params(params)
        single_entry = isinstance(params, SingleEntryQueryParams)
        response_fields = criteria.pop("fields")

        results, data_returned, more_data_available = self._run_db_query(
            criteria, single_entry=isinstance(params, SingleEntryQueryParams)
        )

        results = self._es_to_optimade_results(results, response_fields=response_fields)

        if single_entry:
            results = results[0] if results else None

            if data_returned > 1:
                raise HTTPException(
                    status_code=404,
                    detail=f'Instead of a single entry, {data_returned} entries were found')

        exclude_fields = self.all_fields - response_fields

        return (
            results,
            data_returned,
            more_data_available,
            exclude_fields,
        )

    def _check_aliases(self, aliases):
        pass

    def _es_to_optimade_result(
            self, es_result: dict,
            response_fields: Set[str],
            upload_files_cache: Dict[str, files.UploadFiles]) -> StructureResource:

        calc_id, upload_id = es_result['calc_id'], es_result['upload_id']
        upload_files = upload_files_cache.get(upload_id)

        if upload_files is None:
            upload_files = files.UploadFiles.get(upload_id, is_authorized=lambda: True)
            if upload_files is None:
                logger.error('missing upload', upload_id=upload_id)
                return None

            upload_files_cache[upload_id] = upload_files

        try:
            archive = upload_files.read_archive(calc_id)
        except KeyError:
            logger.error('missing archive entry', upload_id=upload_id, calc_id=calc_id)
            return None

        metadata = archive[calc_id]['section_metadata'].to_dict()
        entry = datamodel.EntryMetadata.m_from_dict(metadata)

        attrs = entry.dft.optimade.m_to_dict()

        attrs['immutable_id'] = calc_id
        attrs['last_modified'] = entry.last_processing if entry.last_processing is not None else entry.upload_time

        # TODO this should be removed, once all data is reprocessed with the right normalization
        attrs['chemical_formula_reduced'] = optimade_chemical_formula_reduced(
            attrs['chemical_formula_reduced'])
        attrs['chemical_formula_anonymous'] = optimade_chemical_formula_anonymous(
            attrs['chemical_formula_reduced'])
        attrs['chemical_formula_hill'] = optimade_chemical_formula_hill(
            attrs['chemical_formula_hill'])
        attrs['chemical_formula_descriptive'] = attrs['chemical_formula_hill']
        dimension_types = attrs['dimension_types']
        if isinstance(dimension_types, int):
            attrs['dimension_types'] = [1] * dimension_types + [0] * (3 - dimension_types)
            attrs['nperiodic_dimensions'] = dimension_types
        elif isinstance(dimension_types, list):
            attrs['nperiodic_dimensions'] = sum(dimension_types)

        if response_fields is not None:
            for request_field in response_fields:
                if not request_field.startswith('_nmd_'):
                    continue

                if request_field == '_nmd_archive_url':
                    attrs[request_field] = config.api_url() + f'/archive/{upload_id}/{calc_id}'
                    continue

                if request_field == '_nmd_entry_page_url':
                    attrs[request_field] = config.gui_url(f'entry/id/{upload_id}/{calc_id}')
                    continue

                if request_field == '_nmd_raw_file_download_url':
                    attrs[request_field] = config.api_url() + f'/raw/calc/{upload_id}/{calc_id}'
                    continue

                try:
                    if request_field.startswith('_nmd_dft_'):
                        attrs[request_field] = getattr(entry.dft, request_field[9:])
                    else:
                        attrs[request_field] = getattr(entry, request_field[5:])
                except AttributeError:
                    # if unknown properties where provided, we will ignore them
                    pass

        return self.resource_cls(
            type='structures',
            id=entry.calc_id,
            attributes=attrs,
            relationships=None)

    def _es_to_optimade_results(self, es_results: List[dict], response_fields: Set[str]):
        upload_files_cache: Dict[str, files.UploadFiles] = {}
        optimade_results = []
        try:
            for es_result in es_results:
                optimade_result = self._es_to_optimade_result(
                    es_result, response_fields, upload_files_cache)
                if optimade_result is not None:
                    optimade_results.append(optimade_result)
        finally:
            for upload_files in upload_files_cache.values():
                upload_files.close()

        return optimade_results

    def _run_db_query(self, criteria: Dict[str, Any], single_entry=False):

        sort, order = criteria.get('sort', (('chemical_formula_reduced', 1),))[0]
        sort_quantity = datamodel.OptimadeEntry.m_def.all_quantities.get(sort, None)
        if sort_quantity is None:
            raise BadRequest(detail='Unable to sort on field %s' % sort)
        sort_quantity_a_optimade = sort_quantity.m_get_annotations('optimade')
        if not sort_quantity_a_optimade.sortable:
            raise BadRequest(detail='Unable to sort on field %s' % sort)

        search_request = self._base_search_request().include('calc_id', 'upload_id')

        if criteria.get("filter", False):
            search_request.query(criteria["filter"])

        es_response = search_request.execute_paginated(
            page_offset=criteria.get('skip', 0),
            per_page=criteria['limit'],
            order=order,
            order_by='dft.optimade.%s' % sort)
        results = es_response['results']

        data_returned = es_response['pagination']['total']
        more_data_available = data_returned >= criteria.get('skip', 0) + criteria['limit']

        return results, data_returned, more_data_available

    def insert(self, *args, **kwargs):
        # This is used to insert test records during OPT tests. This should never be necessary
        # on our implementation. We just need to implement it, because its marked as
        # @abstractmethod.
        raise NotImplementedError()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import re
from datetime import datetime

from nomad.units import ureg
from nomad.parsing.parser import FairdiParser
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.common_dft import Run, Method, XCFunctionals, System,\
    BasisSetCellDependent, SingleConfigurationCalculation, SamplingMethod, ScfIteration,\
    Eigenvalues, KBand, KBandSegment, Topology, AtomType

from castepparser.metainfo import m_env
from castepparser.metainfo.castep import x_castep_section_phonons, x_castep_section_scf_parameters,\
    x_castep_section_density_mixing_parameters, x_castep_section_population_analysis_parameters,\
    x_castep_section_core_parameters, x_castep_section_band_parameters,\
    x_castep_section_ts_parameters, x_castep_section_optics_parameters,\
    x_castep_section_electronic_spectroscpy_parameters, x_castep_section_tddft_parameters,\
    x_castep_section_atom_positions, x_castep_section_population_analysis,\
    x_castep_section_vibrational_frequencies, x_castep_section_tddft, x_castep_section_DFT_SEDC,\
    x_castep_section_van_der_Waals_parameters, x_castep_section_time, x_castep_section_raman_tensor


# TODO map all castep units
units_map = {
    'A': ureg.angstrom, 'amu': ureg.amu, 'ps': ureg.ps, 'e': ureg.e, 'eV': ureg.eV,
    'K': ureg.K, 'GPa': ureg.GPa, 'hbar': ureg.hbar, 'cm': ureg.cm, 'D': ureg.D,
    'mol': ureg.mol, 'J': ureg.J, 'atom': 1, 'steps': 1, 'iterations': 1}


def resolve_unit(unit_str, parts=[]):
    unit_str = unit_str.replace(' ', '')
    parts = list(parts)

    if unit_str in units_map:
        return units_map[unit_str]

    try:
        return float(unit_str)
    except Exception:
        pass

    if unit_str == '':
        return 1

    open_p = unit_str.rfind('(')
    if open_p > -1:
        n_groups = unit_str.count('(')
        if n_groups != unit_str.count(')'):
            return unit_str
        for n in range(n_groups):
            part = unit_str[open_p + 1:]
            part = part[:part.find(')')]
            parts.append(resolve_unit(part, parts))
            unit_str = unit_str.replace('(%s)' % part, '[%d]' % n)
            open_p = unit_str.rfind('(')
        return resolve_unit(unit_str, parts)

    vals = unit_str.split('/')
    if len(vals) > 1:
        vals = [resolve_unit(v, parts) for v in vals]
        val = vals[0]
        for v in vals[1:]:
            val /= v
        return val

    vals = unit_str.split('**')
    if len(vals) > 1:
        vals = [resolve_unit(v, parts) for v in vals]
        val = vals[0]
        for v in reversed(vals[1:]):
            val = val ** v
        return val

    vals = unit_str.split('*')
    if len(vals) > 1:
        vals = [resolve_unit(v, parts) for v in vals]
        unit = 1
        for v in vals:
            unit *= v
        return unit

    vals = unit_str.split('-1')
    if len(vals) == 2:
        return 1 / resolve_unit(vals[0], parts)

    vals = re.match(r'\[(\d+)\]', unit_str)
    if vals:
        return parts[int(vals.group(1))]


class CellParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_block(val_in):
            val = val_in.split('\n')
            key = val[0].strip().lower()
            return key, [v.split() for v in val[1:] if v and v[0] not in ('#', '!')]

        def str_to_value(val_in):
            key, val = val_in.split(' ', 1)
            return key.strip().lower(), val

        self._quantities = [
            Quantity(
                'block',
                r'\s\%(?:block|BLOCK)\s*([\s\S]+?)\%(?:endblock|ENDBLOCK)',
                str_operation=str_to_block, repeats=True),
            Quantity(
                'value',
                r'\s(\w.+?)\s*:\s*([^\n]+)', str_operation=str_to_value, repeats=True)]

    def get_value(self, key, default=None):
        for source in ['value', 'block']:
            for val in self.get(source, []):
                if val[0] == key:
                    return val[1]
        return default


class BandsParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_kpt_energies(val_in):
            val = [v.split() for v in val_in.split('\n')]
            kpt = np.array(val[0][:3], dtype=np.dtype(np.float64))
            energies = np.array([v[0] for v in val if len(v) == 1], dtype=np.dtype(np.float64))
            return kpt, energies

        self._quantities = [
            Quantity('n_kpoints', r'Number of k-points\s*(\d+)', dtype=np.int32),
            Quantity('n_spins', r'Number of spin components\s*(\d+)', dtype=np.int32),
            Quantity('n_electrons', r'Number of electrons\s*([\d\. ]+)', dtype=np.float64),
            Quantity('n_eigenvalues', r'Number of eigenvalues\s*([\d\. ]+)', dtype=np.float64),
            Quantity(
                'fermi_energies',
                r'Fermi energies \(in atomic units\)\s*([\d\.\- ]+)',
                dtype=np.float64, units=ureg.hartree),
            Quantity(
                'kpt_energies',
                r'point\s*\d+\s*([\s\S]+?)(?:K\-|\Z)',
                repeats=True, str_operation=str_to_kpt_energies, convert=False)]


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_float = r'[\d\.\-\+Ee]+'

        def str_to_title(val_in):
            def add_unit(val):
                val_unit = val.rsplit(' ', 1)
                if len(val_unit) == 2:
                    try:
                        unit = resolve_unit(val_unit[1].strip())
                        val = float(val_unit[0]) * unit
                    except Exception:
                        pass
                return val

            re_section = re.compile(r' *\*+ *(.+?) *\*+ *')
            re_value = re.compile(r' *(.+?)\s*:\s*(.+)')
            title = dict()
            # necessary to cache last quantity for multi-line values such as custom xc func
            last_parameter = []
            for line in val_in.split('\n'):
                section = re_section.match(line)
                if section:
                    key = section.group(1).lower()
                    title[key] = dict()
                    continue
                value = re_value.match(line)
                if value:
                    sub_key = value.group(1).strip().lower()
                    if sub_key not in title[key]:
                        title[key][sub_key] = []
                    val = value.group(2).strip().strip(':')  # misformattig of custom xc func
                    if not val:
                        last_parameter = [sub_key, []]
                        continue
                    title[key][sub_key].append(add_unit(val))
                    if last_parameter:
                        title[key][last_parameter[0]].extend(last_parameter[1])
                        last_parameter = []
                elif line and last_parameter:
                    last_parameter[1].append(add_unit(line.strip()))

            for key in title.keys():
                for sub_key, val in title[key].items():
                    title[key][sub_key] = val[0] if len(val) == 1 else val

            return title

        def str_to_lattice_vectors(val_in):
            val = val_in.strip().split('\n')
            val = np.array([v.split() for v in val[1:]], dtype=np.dtype(np.float64)).T
            return [val[0:3].T, val[3:6].T]

        def str_to_lattice_parameters(val_in):
            re_parameter = rf'(\w+)\s*=\s*({re_float})'
            parameters = {p[0]: float(p[1]) for p in re.findall(re_parameter, val_in)}
            return parameters

        def str_to_array(val_in):
            val = [v.strip().strip('x').strip('*').split() for v in val_in.split('\n')]
            val = np.transpose([v for v in val if len(v) == 5])
            array = np.array(val[-3:], dtype=np.dtype(np.float64)).T
            return val[0], array

        def str_to_kpt_energies(val_in):
            spin = re.search(r'Spin=(\d+)', val_in)
            spin = int(spin.group(1)) if spin else 1
            kpt = re.search(r'kpt=\s*\d+\s*\(([\d\.\- ]+)\)', val_in)
            kpt = np.array(kpt.group(1).split(), dtype=np.dtype(np.float64)) if kpt else [0., 0., 0.]
            re_energies = re.compile(rf'\+\s*\d+\s*({re_float})\s*\+')
            energies = re.findall(re_energies, val_in)
            if energies:
                energies = np.array(energies, dtype=np.dtype(np.float64))
            return spin, kpt, energies

        def str_to_scf(val_in):
            result = []
            for val in val_in.strip().split('\n'):
                val = val.split()
                if not val:
                    continue
                if val[0].isdecimal() or val[0] == 'Initial':
                    val = val[:-2]
                    if val[0] == 'Initial':
                        val.insert(-1, 0.0)
                    result.append(val[1:])
            return np.array(result, dtype=np.dtype(np.float64))

        def str_to_md_data(val_in):
            re_data = re.compile(rf'x\s*(.+?):\s*({re_float})\s*(\w+)')
            data = dict()
            for val in val_in.split('\n'):
                val = re_data.search(val)
                if val:
                    data[val.group(1).strip()] = float(val.group(2)) * units_map.get(val.group(3), 1)
            return data

        def str_to_energy(val_in):
            key, val = val_in.split('=', 1)
            val = val.split()
            try:
                unit = resolve_unit(val[1])
                unit = unit if unit else 1
            except Exception:
                unit = 1
            return key.strip(), float(val[0]) * unit

        def str_to_mulliken(val_in):
            val = [v.split() for v in val_in.split('\n')]
            keys = [v for v in val[0] if not v.startswith('(')]
            return {key: [v[n] for v in val[1:] if len(v) == len(keys)] for n, key in enumerate(keys)}

        def str_to_vibrational_frequencies(val_in):
            # replace keys to fit length of values
            val_in = re.sub(r'ir\s*intensity\s*active', 'ir_intensity ir_active', val_in)
            val_in = re.sub(r'raman\s*activity\s*active', 'raman_activity raman_active', val_in)
            val_in = re.sub(r'raman\s*active', 'raman_active', val_in)
            val_in = re.sub(r'Frequency', 'vibrational_frequencies', val_in)
            val = [v.strip(' +').split() for v in val_in.split('\n')]
            return {key: [v[n] for v in val[1:] if len(v) == len(val[0])] for n, key in enumerate(val[0])}

        system_quantities = [
            Quantity(
                'unit_cell',
                r'Unit Cell\s*\s*\-+([\s\S]+?)\-{20}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'lattice_vectors',
                        r'(Real.+\s*[\d\.\-\s]+)',
                        str_operation=str_to_lattice_vectors, convert=False),
                    Quantity(
                        'lattice_parameters',
                        r'(Lattice parameters.+[\s\S]+?\n *\n)',
                        str_operation=str_to_lattice_parameters, convert=False),
                    Quantity(
                        'cell_volume',
                        rf'Current cell volume\s*=\s*({re_float})')])),
            Quantity(
                'cell_contents',
                r'Cell Contents\s*\-+([\s\S]+?)\n *\n *\-{20}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'positions',
                        r'u\s*v\s*w\s*x([\s\S]+?)x{50}',
                        str_operation=str_to_array, convert=False),
                    Quantity(
                        'velocities',
                        r'Vx\s*Vy\s*Vz\s*x([\s\S]+?)x{50}',
                        str_operation=str_to_array, convert=False)])),
            Quantity(
                'species',
                r'Details of Species\s*\-+([\s\S]+?)\-{20}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'mass',
                        r'Mass of species in AMU\s*([\s\S]+?)\n *\n',
                        str_operation=lambda x: [v.split() for v in x.strip().split('\n')])])),
            Quantity(
                'dft_d',
                r'(Dispersion-correction scheme[\s\S]+?)\n *\n *\n',
                sub_parser=TextParser(quantities=[
                    Quantity('method', r'Dispersion\-correction scheme\s*\:\s*(\w+)'),
                    Quantity(
                        'parameter',
                        rf'Parameter\s*(\w+)\s*\:\s*({re_float})', repeats=True)]))]

        basic_quantities = [
            Quantity(
                'scf',
                r'(SCF loop.+\s*.+\s*\-+ \<\-\- SCF[\s\S]+?)\-{70}\s*\<\-\- SCF',
                str_operation=str_to_scf, convert=False),
            Quantity(
                'energy',
                rf'(.+?energy.+?\s*=\s*{re_float}\s*.+)\n',
                repeats=True, str_operation=str_to_energy, convert=False),
            Quantity('energy_total', rf'Final Total Energy\s*({re_float})', dtype=float),
            Quantity(
                'enthalpy',
                rf'(Final Enthalpy\s*=\s*{re_float}\s*.+)',
                str_operation=str_to_energy, convert=False),
            Quantity(
                'frequency',
                rf'(Final \<frequency\>\s*=\s*{re_float}\s*.+)',
                str_operation=str_to_energy, convert=False),
            Quantity(
                'forces',
                r'Forces \*+\s*([\s\S]+?)\*{20}',
                str_operation=str_to_array, convert=False),
            Quantity(
                'stress_tensor',
                r'Stress Tensor \*+\s*([\s\S]+?)\*{20}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'stress_tensor',
                        rf'\*\s*(?:x|y|z)\s*({re_float})\s*({re_float})\s*({re_float})',
                        repeats=True, dtype=np.dtype(np.float64)),
                    Quantity(
                        'pressure',
                        rf'\*\s*Pressure:\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'mulliken',
                r'Atomic Populations \(Mulliken\)\s*\-+\s*([\s\S]+?)\n *\n',
                str_operation=str_to_mulliken),
            Quantity(
                'tddft',
                r'Time-Dependent DFT Calculation([\s\S]+?TDDFT calculation time.+)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        r'TDDFT iteration\:\s*(\d+).+?Time\:\s*([\d\.]+)', repeats=True),
                    Quantity(
                        'energies',
                        rf'(\d+)\s*({re_float})\s*({re_float})\s*\w+',
                        repeats=True, dtype=np.dtype(np.float64)),
                    Quantity(
                        'time',
                        r'TDDFT calculation time\:\s*([\d\.]+)')])),
            Quantity(
                'interaction_energy',
                r'SEDC PBC Interaction Energy([\s\S]+?NB dispersion corrected est.+)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy',
                        rf'\%\s*(.+?=\s*{re_float}\s*)\[(.+)\]',
                        repeats=True, str_operation=str_to_energy, convert=False),
                    Quantity(
                        'shell',
                        rf'\%\s*(\d+\s+{re_float}\s+{re_float}\s+{re_float})', repeats=True)]))]

        bandstructure_quantities = [
            Quantity(
                'fermi_energy',
                rf'Fermi energy for spin\s*(?:up|down)\s*electrons is:\s*({re_float})',
                repeats=True, dtype=np.float64),
            Quantity(
                'fermi_energy',
                rf'Fermi energy for spin-degenerate system:\s*({re_float})',
                repeats=True, dtype=np.float64),
            Quantity(
                'kpt_energies',
                r'(Spin=\d+ kpt=.+\s*\+ *\-+ *\+[\s\S]+?)(?:\-{50}|\Z)',
                repeats=True, str_operation=str_to_kpt_energies)]

        basis_set_correction_quantities = [
            Quantity(
                'iteration',
                r'(ulating total energy[\s\S]+?)(?:Calc|\Z)',
                repeats=True, sub_parser=TextParser(quantities=basic_quantities + [
                    Quantity(
                        'cutoff',
                        rf'Calculating total energy with cut-off of\s*([\d\.]+)',
                        dtype=np.float64)]))]

        md_quantities = basic_quantities + [
            Quantity(
                'md_data',
                r'MD Data:\s*x([\s\S]+?)x{50}',
                str_operation=str_to_md_data, convert=False),
            Quantity(
                'iteration',
                r'(Starting MD iteration\s*\d+[\s\S]+?)finished',
                repeats=True, sub_parser=TextParser(quantities=basic_quantities + system_quantities))]

        dmd_quantities = basic_quantities + [
            Quantity(
                'iteration',
                r'(Starting DMD iteration\s*\d+[\s\S]+?)finished',
                repeats=True, sub_parser=TextParser(quantities=basic_quantities + system_quantities))]

        di_quantities = basic_quantities + [
            Quantity(
                'iteration',
                r'(Starting DI iteration\s*\d+[\s\S]+?)finished',
                repeats=True, sub_parser=TextParser(quantities=basic_quantities + system_quantities))]

        tss_quantities = basic_quantities + [
            Quantity(
                'iteration',
                r'(SCF loop\s*Energy[\s\S]+?Energy\:.+)',
                repeats=True, sub_parser=TextParser(quantities=basic_quantities))]

        cg_refinement_quantities = basic_quantities + [
            Quantity(
                'iteration',
                r'(SCF loop\s*Energy[\s\S]+?NB est\. 0K energy.+)',
                repeats=True, sub_parser=TextParser(quantities=basic_quantities))]

        bfgs_iteration_quantities = basic_quantities + system_quantities + [
            Quantity(
                'spin_density',
                rf'Integrated Spin Density\s*=\s*({re_float})')]

        bfgs_quantities = basic_quantities + [
            Quantity(
                'iteration',
                r'(Starting BFGS iteration\s*\d+.+\s*\=+[\s\S]+?)\={70}',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        r'((?:starting|improving) iteration[\s\S]+?)BFGS:',
                        repeats=True, sub_parser=TextParser(quantities=bfgs_iteration_quantities))]))]

        self._quantities = system_quantities + [
            Quantity(
                'program_version',
                r'Welcome to (?:Academic Release|Materials Studio)\s*(\w*) version (\S+)'),
            Quantity(
                'program_compilation',
                r'Compiled for (\S+) on \w+,\s*([\w :]+)', flatten=False),
            Quantity('compiler', r'Compiler:\s*(.+)\n', flatten=False),
            Quantity('maths_library', r'MATHLIBS\s*:\s*(.+)\n', flatten=False),
            Quantity('fft_library', r'FFT Lib\s*:\s*(.+)\n', flatten=False),
            Quantity(
                'constants_reference',
                r'Fundamental constants values\s*:\s*(.+)\n', flatten=False),
            Quantity('run_start', r'Run started\: \w+,\s*([\w :]+)', flatten=False),
            Quantity('title', r'Title\s*\*+\s*([\s\S]+?)\*{75}', str_operation=str_to_title),
            # TODO verify this, I cannot find an example
            Quantity('dft_u', r'Units for (\w+) U values are'),
            Quantity(
                'calculation',
                r'(MEMORY AND SCRATCH DISK ESTIMATES PER[\s\S]+?(?:A BibTeX formatted|\Z))',
                sub_parser=TextParser(quantities=basic_quantities + [
                    Quantity(
                        'vibrational_frequencies',
                        r'\+\s*(N\s*Frequency\s*irrep[\s\S]+?)(?:\s*\+\s*\.{70}|\Z)',
                        repeats=True, str_operation=str_to_vibrational_frequencies),
                    Quantity(
                        'raman_tensor',
                        rf'Mode number\:\s*\d+ Raman tensor\s*Depolarisation Ratio.+'
                        rf'\s*\+\s*({re_float}\s*{re_float}\s*{re_float}).+'
                        rf'\s*\+\s*({re_float}\s*{re_float}\s*{re_float}).+'
                        rf'\s*\+\s*({re_float}\s*{re_float}\s*{re_float}).+',
                        repeats=True, str_operation=lambda x: np.reshape(
                            np.array(x.split(), dtype=np.dtype(np.float64)), (3, 3))),
                    Quantity(
                        'bandstructure',
                        r'(\={50}[\s\S]+?B A N D   S T R U C T U R E   C A L C U L A T I O N)'
                        r'([\s\S]+?Spin\=1[\s\S]+?)(?:\={50}|\Z)',
                        sub_parser=TextParser(quantities=bandstructure_quantities)),
                    Quantity(
                        'basis_set_correction',
                        r'(Calculating finite basis set correction with[\s\S]+?)'
                        r'(Total energy corrected for finite basis set.+)',
                        sub_parser=TextParser(quantities=basis_set_correction_quantities)),
                    Quantity(
                        'tss',
                        r'Starting Transition State Search([\s\S]+?[LQ]ST Maximum Found[\s\S]+?\+{10})',
                        sub_parser=TextParser(quantities=tss_quantities)),
                    Quantity(
                        'cg_refinement',
                        r'Commencing conjugate gradient refinement([\s\S]+?)(?:Writing model|\Z)',
                        sub_parser=TextParser(quantities=cg_refinement_quantities)),
                    Quantity(
                        'md',
                        r'Starting MD([\s\S]+?)(?:Finished MD|\Z)',
                        sub_parser=TextParser(quantities=md_quantities)),
                    Quantity(
                        'dmd',
                        r'Starting DMD([\s\S]+?)(?:Finished DMD|\Z)',
                        sub_parser=TextParser(quantities=dmd_quantities)),
                    # TODO verify this cannot find an example
                    Quantity(
                        'di',
                        r'Starting DI([\s\S]+?)(?:Finished DI|\Z)',
                        sub_parser=TextParser(quantities=di_quantities)),
                    Quantity(
                        'bfgs',
                        r'(BFGS:[\s\S]+?)(?:BFGS: Geometry optimization|\Z)',
                        sub_parser=TextParser(quantities=bfgs_quantities)),
                    Quantity(
                        'final',
                        r'Final Configuration([\s\S]+?)(?:Writing model to|\Z)',
                        sub_parser=TextParser(quantities=system_quantities + basic_quantities))])),
            Quantity(
                'time',
                r'([A-Z]\w+) time\s+\=\s*([\d\.]+) s',
                dtype=np.float64, repeats=True)]


class CastepParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/castep', code_name='CASTEP', code_homepage='http://www.castep.org/',
            mainfile_contents_re=(r'\s\|\s*CCC\s*AA\s*SSS\s*TTTTT\s*EEEEE\s*PPPP\s*\|\s*'))
        self._metainfo_env = m_env
        self.out_parser = OutParser()
        self.cell_parser = CellParser()
        self.bands_parser = BandsParser()

        self._xc_functional_map = {
            'Perdew Burke Ernzerhof': ['GGA_X_PBE', 'GGA_C_PBE'],
            'Local Density Approximation': ['LDA_C_PZ_LDA_X_PZ'],
            'Perdew Wang (1991)': ['GGA_X_PW91', 'GGA_C_PW91'],
            'revised Perdew Burke Ernzerhof': ['GGA_X_RPBE'],
            'PBE for solids (2008)': ['GGA_X_PBE_SOL'],
            'hybrid B3LYP': ['HYB_GGA_XC_B3LYP5'],
            'Hartree-Fock': ['HF_X'],
            'Hartree-Fock + Local Density Approximation': ['HF_X_LDA_C_PW'],
            'hybrid HSE03': ['HYB_GGA_XC_HSE03'],
            'hybrid HSE06': ['HYB_GGA_XC_HSE06'],
            'hybrid PBE0': ['HYB_GGA_XC_PBEH'],
            'PBE with Wu-Cohen exchange': ['GGA_C_PBE_GGA_X_WC'],
            'LDA-X': ['LDA_X_PZ'],
            'LDA-C': ['LDA_C_PZ'],
            'Optimised Effective Potential': ['OEP_EXX'],
            'PBE': ['GGA_X_PBE', 'GGA_C_PBE'],
            'PW91': ['GGA_X_PW91', 'GGA_C_PW91'],
            'RPBE': 'GGA_X_RPBE',
            'PBEsol': ['GGA_X_RPBE'],
            'HF': ['HF_X'],
            'HF-LDA': ['HF_X_LDA_C_PW'],
            'HSE03': ['HYB_GGA_XC_HSE03'],
            'HSE06': ['HYB_GGA_XC_HSE06'],
            'PBE0': ['HYB_GGA_XC_PBEH'],
            'WC': ['GGA_C_PBE_GGA_X_WC'],
            'LDA': ['LDA_X_PZ', 'LDA_C_PZ'],
            'b3lyp': ['HYB_GGA_XC_B3LYP5']}

        self._relativistic_map = {
            'Koelling-Harmon': 'scalar_relativistic'}

        self._sampling_method_map = {
            'single point energy': 'single_point',
            'geometry optimization': 'geometry_optimization',
            'molecular dynamics': 'molecular_dynamics',
            'phonon calculation': 'phonon',
            'transition state search': 'geometry_optimization',
            'band structure': 'single_point',
            'Core level spectra (ELNES etc)': 'single_point'}

        self._metainfo_map = {
            'Final energy': 'energy_total', 'Final energy, E': 'energy_total',
            'Total energy corrected for finite basis set': 'x_castep_total_energy_corrected_for_finite_basis',
            'Final free energy (E-TS)': 'energy_free',
            'Dispersion corrected final energy*': 'x_castep_total_dispersion_corrected_energy',
            'NB est. 0K energy (E-0.5TS)': 'energy_total_T0',
            "'TS'/PBE structure energy corr.'": 'x_castep_structure_energy_corr',
            "'TS'/PBE PBC image interaction corr.": 'x_castep_PBC_image_inter_corr',
            "'TS'/PBE total energy correction": 'x_castep_total_energy_correction',
            "'TS'/PBE correction |F|max": 'x_castep_total_fmax_correction',
            'Dispersion corrected final energy*, Ecor': 'x_castep_total_dispersion_corrected_energy',
            'Dispersion corrected final free energy* (Ecor-TS)': 'x_castep_total_dispersion_corrected_free_energy',
            'NB dispersion corrected est. 0K energy* (Ecor-0.5TS)': 'x_castep_disp_corrected_energy_total_T0',
            'basis set parameters': {
                'size of standard grid': 'x_castep_size_std_grid',
                'size of   fine   gmax': 'x_castep_size_fine_grid'},
            'phonon parameters': {
                'phonon calculation method': 'x_castep_phonon_method',
                'phonon convergence tolerance': 'x_castep_phonon_tolerance',
                'max. number of phonon cycles': 'x_castep_phonon_cycles',
                'dfpt solver method': 'x_castep_DFPT_solver_method',
                'band convergence tolerance': 'x_castep_band_tolerance'},
            'geometry optimization parameters': {
                'optimization method': 'geometry_optimization_method',
                'total energy convergence tolerance': 'geometry_optimization_energy_change',
                'max. number of steps': 'x_castep_max_number_of_steps',
                'max ionic |force| tolerance': 'geometry_optimization_threshold_force',
                'max ionic |displacement| tolerance': 'geometry_optimization_geometry_change',
                'max |stress component| tolerance': 'x_castep_geometry_stress_com_tolerance'},
            'electronic parameters': {
                'number of  electrons': 'x_castep_number_of_electrons',
                'net charge of system': 'x_castep_net_charge',
                'number of bands': 'x_castep_number_of_bands'},
            'electronic minimization parameters': {
                'total energy / atom convergence tol.': 'x_castep_energy_threshold',
                'max. number of scf cycles': 'x_castep_max_iter',
                'smearing scheme': 'x_castep_smearing_kind',
                'smearing width': 'x_castep_smearing_width'},
            'density mixing parameters': {
                'density-mixing scheme': 'x_castep_density_mixing_scheme',
                'max. length of mixing history': 'x_castep_density_mixing_length',
                'charge density mixing amplitude': 'x_castep_charge_density_mixing_amplitude',
                'cut-off energy for mixing': 'x_castep_cut_off_energy_for_mixing'},
            'population analysis parameters': {
                'population analysis with cutoff': 'x_castep_population_analysis_cutoff'},
            'core level spectra parameters': {
                'number of bands': 'x_castep_core_spectra_n_bands',
                'band convergence tolerance': 'x_castep_core_spectra_conv_tolerance'},
            'band structure parameters': {
                'max. number of iterations': 'x_castep_band_n_iterations',
                'max. cg steps in bs calc': 'x_castep_band_max_cg',
                'number of bands / k-point': 'x_castep_band_n_bands',
                'band convergence tolerance': 'x_castep_band_conv_tolerance'},
            'molecular dynamics parameters': {
                'ensemble': 'ensemble_type',
                'temperature': 'x_castep_thermostat_target_temperature',
                'pressure': 'x_castep_frame_pressure',
                'using': ['x_castep_barostat_type', 'x_castep_thermostat_type'],
                'with characteristic cell time': 'x_castep_barostat_tau',
                'with characteristic ionic time': 'x_castep_thermostat_tau',
                'time step': 'x_castep_integrator_dt',
                'number of md steps': 'x_castep_number_of_steps_requested',
                'md scf energy / atom convergence tol.': 'x_castep_frame_energy_tolerance',
                'md scf eigenenergies tolerance': 'x_castep_frame_eigen_tolerance'},
            'transition state search parameters': {
                'search method': 'x_castep_ts_method',
                'lstqst protocol': 'x_castep_ts_protocol',
                'max. number of qst iterations': 'x_castep_ts_number_qst',
                'max. number of cg iterations': 'x_castep_ts_number_cg',
                'max. ionic |force| tolerance': 'x_castep_ts_force_tolerance',
                'max. ionic |displacement| tolerance': 'x_castep_ts_displacement_tolerance'},
            'optics parameters': {
                'search method': 'x_castep_optics_n_bands',
                'band convergence tolerance': 'x_castep_optics_tolerance'},
            'electronic spectroscopy parameters': {
                'electronic spectroscopy with theory level': 'x_castep_theory_level',
                'spectroscopy calculation': 'x_castep_spectroscopy_calculation',
                'max. number of iterations': 'x_castep_spec_max_iter',
                'max. steps per iteration': 'x_castep_spec_max_steps',
                'number of bands / k-point': 'x_castep_spec_max_bands',
                'band convergence tolerance': 'x_castep_spec_tolerance'},
            'time-dependent dft parameters': {
                'number of excited states': 'x_castep_tddft_n_excited_states',
                'state selected for calculation of forces': 'x_castep_tddft_n_states_forces',
                'state convergence tolerance': 'x_castep_tddft_state_tolerance',
                'convergence tolerance window': 'x_castep_tddft_state_tolerance_window',
                'max. number of iterations': 'x_castep_tddft_max_iter',
                'no. of extra (convergence indifferent) states': 'x_castep_tddft_extra_states',
                'using tddft functional': 'x_castep_tddft_functional',
                'time-dependent dft method': 'x_castep_tddft_method',
                'matrix eigenvalue method': 'x_castep_tddft_eigenmethod',
                'time-Dependent dft approximation': 'x_castep_tddft_approximation',
                'time-Dependent dft position operator': 'x_castep_tddft_position_op'}}

        self._units = None
        self._nspin = None

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger
        self._units = None
        self._nspin = None

    @property
    def units(self):
        if self._units is None:
            self._units = dict()
            re_unit = re.compile(r'output\s*([\w ]+) unit')
            parameters = self.out_parser.get('title', {}).get('general parameters', {})
            for key, val in parameters.items():
                name = re_unit.match(key)
                if name:
                    try:
                        self._units[name.group(1)] = resolve_unit(val)
                    except Exception:
                        pass
        return self._units

    @property
    def n_spin_channels(self):
        if self._nspin is None:
            self._nspin = 1 if self.out_parser.get('title', {}).get(
                'electronic parameters', {}).get('number of down spins') is None else 2
        return self._nspin

    def get_castep_file(self, ext):
        paths = [p for p in os.listdir(self.maindir) if p.endswith(ext)]
        if not paths:
            return
        elif len(paths) == 1:
            return os.path.join(self.maindir, paths[0])
        else:
            prefix = os.path.basename(self.filepath).rsplit('.', 1)[0]
            for path in paths:
                if path.startswith(prefix):
                    return os.path.join(self.maindir, path)

    def parse_method(self):
        sec_method = self.archive.section_run[0].m_create(Method)
        title = self.out_parser.get('title', {})

        method = 'DFT+U' if self.out_parser.get('dft_u') is not None else 'DFT'
        sec_method.electronic_structure_method = method
        sec_method.number_of_spin_channels = self.n_spin_channels

        xc_parameters = title.get('exchange-correlation parameters', {})

        # xc functional
        xc_functionals = xc_parameters.get('using functional')
        if xc_functionals is None:
            xc_functionals = xc_parameters.get('using custom xc functional definition', [])
        xc_functionals = xc_functionals if isinstance(xc_functionals, list) else [xc_functionals]

        for xc_functional in xc_functionals:
            # when custom, a weight is attached
            xc_weight = None
            if xc_parameters.get('using custom xc functional definition') is not None:
                xc_functional, xc_weight = xc_functional.rsplit(' ', 1)
            for functional in self._xc_functional_map.get(xc_functional, []):
                sec_xc_functional = sec_method.m_create(XCFunctionals)
                sec_xc_functional.XC_functional_name = functional
                if xc_weight is not None:
                    sec_xc_functional.XC_functional_weight = float(xc_weight)

        # relativistic treatment
        relativistic = self._relativistic_map.get(xc_parameters.get('relativistic treatment'))
        if relativistic is not None:
            sec_method.relativity_method = relativistic

        # dispersion correction
        dispersion = xc_parameters.get('sedc with')
        if dispersion is not None:
            sec_method.van_der_Waals_method = dispersion.split()[0]

        # smearing scheme
        electronic_parameters = title.get('electronic minimization parameters', {})
        smearing = electronic_parameters.get('smearing scheme')
        if smearing is not None:
            sec_method.smearing_kind = smearing.lower()
            width = electronic_parameters.get('smearing width')
            if width is not None:
                sec_method.smearing_width = width.to('joule').magnitude

    def parse_sampling_method(self):
        sec_run = self.archive.section_run[0]
        sec_sampling_method = sec_run.section_sampling_method
        sec_sampling_method = sec_sampling_method[-1] if sec_sampling_method else sec_run.m_create(SamplingMethod)

        title = self.out_parser.get('title', {})

        method = title.get('general parameters', {}).get('type of calculation', 'single point')
        method = self._sampling_method_map.get(method)
        if method is None:
            self.logger.warn('Sampling method cannot be resolved.')
            method = 'single_point'
        sec_sampling_method.sampling_method = method

        if method == 'geometry_optimization':
            for key, val in title.get('geometry optimization parameters', {}).items():
                key = self._metainfo_map.get('geometry optimization parameters').get(key)
                if key is None or val is None:
                    continue
                setattr(sec_sampling_method, key, val)
        elif method == 'molecular_dynamics':
            for key, val in title.get('molecular dynamics parameters', {}).items():
                key = self._metainfo_map.get('molecular dynamics parameters').get(key)
                if key is None or val is None:
                    continue
                if isinstance(key, list):
                    for i, key_i in enumerate(key):
                        setattr(sec_sampling_method, key_i, val[i])
                else:
                    setattr(sec_sampling_method, key, val)

    def parse_topology(self):
        species = self.out_parser.get('species')
        if species is None:
            return

        sec_topology = self.archive.section_run[0].m_create(Topology)
        for name, mass in species.get('mass', []):
            sec_atom_type = sec_topology.m_create(AtomType)
            sec_atom_type.atom_type_name = name
            sec_atom_type.atom_type_mass = mass

    def parse_configurations(self):
        calculation = self.out_parser.get('calculation')
        if calculation is None:
            return

        sec_run = self.archive.section_run[0]

        def parse_eigenvalues(source):
            sec_scc = sec_run.section_single_configuration_calculation[-1]

            energy_unit = self.units.get('energy', 1)
            bandstructure = source.get('bandstructure')
            kpts, band_energies = [], []
            if bandstructure is not None:
                for spin, kpt, energies in bandstructure.get('kpt_energies', []):
                    if spin == 1:
                        kpts.append(kpt)
                    band_energies.append(energies)
                band_energies = np.reshape(
                    band_energies, (spin, len(kpts), len(band_energies[0]))) * energy_unit

            else:
                self.bands_parser.mainfile = self.get_castep_file('bands')
                for kpt, energies in self.bands_parser.get('kpt_energies', []):
                    kpts.append(kpt)
                    band_energies.append(energies)
                if len(band_energies) == 0:
                    return

                n_spin = self.bands_parser.get('n_spins', 1)
                band_energies = np.reshape(band_energies, (
                    len(kpts), n_spin, len(band_energies[0]) // n_spin))
                band_energies = np.transpose(band_energies, axes=(1, 0, 2)) * energy_unit

            # get path segment nodes from .cell
            self.cell_parser.mainfile = self.get_castep_file('cell')
            kpoint_path = self.cell_parser.get_value('bs_kpoint_path', [])

            if len(kpoint_path) == 0:
                # write energies to section_eigenvalues
                sec_eigenvalues = sec_scc.m_create(Eigenvalues)
                sec_eigenvalues.eigenvalues_values = band_energies
                sec_eigenvalues.eigenvalues_kpoints = kpts
            else:
                # write band energies on segments
                nodes = np.array([path[:3] for path in kpoint_path], dtype=np.dtype(np.float64))
                labels = ['\u0393' if path[-1].lower() == 'gamma' else path[-1] for path in kpoint_path]
                sec_k_band = sec_scc.m_create(KBand)
                start = 0
                for n, node in enumerate(nodes[1:]):
                    node_index = np.where(kpts == node)[0]
                    for index in node_index:
                        if np.count_nonzero(node_index == index) == 3 and index > start:
                            break
                    sec_band_segment = sec_k_band.m_create(KBandSegment)
                    sec_band_segment.band_energies = band_energies[:, start:index + 1, :]
                    sec_band_segment.band_k_points = kpts[start:index + 1]
                    sec_band_segment.band_segm_labels = [labels[n], labels[n + 1]]
                    start = index

        def parse_scc(source):
            sec_scc = sec_run.m_create(SingleConfigurationCalculation)
            energy_unit = self.units.get('energy', 1)

            if source.get('energy_total') is not None:
                sec_scc.energy_total = source.get('energy_total') * energy_unit

            # energies
            for key, val in source.get('energy', []):
                name = self._metainfo_map.get(key)
                if name is not None:
                    setattr(sec_scc, name, val.to('joule').magnitude)

            # forces
            forces = source.get('forces')
            if forces is not None:
                sec_scc.atom_forces = forces[1] * self.units.get('force', 1)

            # stress tensor
            stress_tensor = source.get('stress_tensor')
            if stress_tensor is not None:
                for key in ['stress_tensor', 'pressure']:
                    val = stress_tensor.get(key)
                    if val is not None:
                        setattr(sec_scc, key, val * self.units.get('pressure', 1))

            # other properties
            for key in ['enthalpy', 'frequency']:
                val = source.get(key, ['', None])[1]
                if val is not None:
                    val = val.to('joule') if key == 'enthalpy' else val
                    setattr(sec_scc, 'x_castep_%s' % key, val.magnitude)

            # eigenvalues
            parse_eigenvalues(source)

            # mulliken population analysis
            mulliken = source.get('mulliken')
            if mulliken is not None:
                # why mulliken section under section_run?
                species = mulliken.get('Species', [])
                for n, specie in enumerate(species):
                    sec_mulliken = sec_run.m_create(x_castep_section_population_analysis)
                    sec_mulliken.x_castep_mulliken_atom = specie
                    sec_mulliken.x_castep_total_orbital = mulliken['Total'][n]
                    sec_mulliken.x_castep_mulliken_charge = mulliken['Charge'][n]
                    for orbital in ('s', 'p', 'd', 'f'):
                        if mulliken.get(orbital) is None:
                            continue
                        setattr(sec_mulliken, 'x_castep_orbital_%s' % orbital, mulliken[orbital][n])

            # vibrational frequencies
            # why are vibrational frequencies section under section_run?
            for vibrational_frequencies in source.get('vibrational_frequencies', []):
                sec_frequencies = sec_run.m_create(x_castep_section_vibrational_frequencies)
                for key, val in vibrational_frequencies.items():
                    setattr(sec_frequencies, 'x_castep_%s' % key, val)

            # raman tensors
            for raman_tensor in source.get('raman_tensor', []):
                sec_raman = sec_run.m_create(x_castep_section_raman_tensor)
                unit = (self.units.get('length', ureg.angstrom) / self.units.get('mass', ureg.amu)) * 0.5
                sec_raman.x_castep_raman_tensor = raman_tensor * unit

            # interaction energy
            interaction = source.get('interaction_energy')
            if interaction is not None:
                for shell in interaction.get('shell', []):
                    sec_sedc = sec_scc.m_create(x_castep_section_DFT_SEDC)
                    sec_sedc.x_castep_shell = shell[0]
                    sec_sedc.x_castep_correction_energy = (shell[1] * energy_unit).to('J').magnitude
                    sec_sedc.x_castep_de_atom = (shell[2] * energy_unit).to('J').magnitude
                    sec_sedc.x_castep_dfmax_atom = (shell[3] * self.units.get('force')).to('J/m').magnitude

                for key, val in interaction.get('energy', []):
                    name = self._metainfo_map.get(key)
                    if name is None:
                        continue
                    val = val.to('J') if val.units == ureg.eV else val.to('J/m')
                    setattr(sec_sedc, name, val.magnitude)

                # add the dispersion energies
                for key, val in source.get('energy', []):
                    name = self._metainfo_map.get(key, '')
                    if '_disp' in name:
                        setattr(sec_sedc, name, val.to('joule').magnitude)

            # scf iteration
            for scf in source.get('scf', []):
                sec_scf = sec_scc.m_create(ScfIteration)
                sec_scf.energy_total_scf_iteration = scf[0] * energy_unit
                if len(scf) == 4:
                    fermi_energy = [scf[1]] * self.n_spin_channels
                    sec_scf.energy_reference_fermi_iteration = fermi_energy * energy_unit
                sec_scf.energy_change_scf_iteration = scf[-2] * energy_unit
                sec_scf.time_scf_iteration = scf[-1]

            return sec_scc

        def parse_system(source):
            sec_system = sec_run.m_create(System)
            length_unit = self.units.get('length', 1)

            unit_cell = source.get('unit_cell', self.out_parser.get('unit_cell', {}))
            if unit_cell.get('lattice_vectors') is not None:
                sec_system.lattice_vectors = unit_cell.get('lattice_vectors')[0] * length_unit
                sec_system.configuration_periodic_dimensions = [True, True, True]

            cell_contents = source.get('cell_contents', self.out_parser.get('cell_contents', {}))
            if cell_contents.get('positions') is not None:
                sec_system.atom_labels = cell_contents.get('positions')[0]
                sec_system.atom_positions = np.dot(cell_contents.get('positions')[1], sec_system.lattice_vectors)
            if cell_contents.get('velocities') is not None:
                sec_system.atom_velocities = cell_contents.get('velocities')[1] * (
                    length_unit / self.units.get('time', 1))

            # miscellaneous quantities
            cell_volume = unit_cell.get('cell_volume')
            if cell_volume is not None:
                cell_volume *= length_unit ** 3
                sec_system.x_castep_cell_volume = cell_volume.to('m**3').magnitude

            lattice_parameters = unit_cell.get('lattice_parameters')
            if lattice_parameters is not None:
                sec_positions = sec_system.m_create(x_castep_section_atom_positions)
                for key, val in lattice_parameters.items():
                    if key in ('a', 'b', 'c'):
                        val *= length_unit
                        setattr(sec_positions, 'x_castep_cell_length_%s' % key, val.to('m').magnitude)
                    elif key in ('alpha', 'beta', 'gamma'):
                        setattr(sec_positions, 'x_castep_cell_angle_%s' % key, val)

            # tddft
            # why are tddft results here?, adapted from old parser
            tddft = source.get('tddft')
            if tddft is not None:
                for energies in tddft.get('energies', []):
                    sec_tddft = sec_system.m_create(x_castep_section_tddft)
                    energies = energies * self.units.get('energy', ureg.eV)
                    sec_tddft.x_castep_state_energy = energies[1].to('J').magnitude
                    sec_tddft.x_castep_state_energy_error = energies[2].to('J').magnitude
                sec_tddft.x_castep_state_number = energies[0].magnitude
                sec_tddft.x_castep_tddft_calculation_time = tddft.get('time', 0.)
                sec_tddft.x_castep_tddft_iteration = tddft.get('iteration', [0, 0])[-1][0]

            # add miscellaneos stuff to first instance
            if len(sec_run.section_system) == 1:
                for key, val in self.out_parser.get('title', {}).get('electronic parameters', {}).items():
                    key = self._metainfo_map.get('electronic parameters', {}).get(key)
                    if key is None:
                        continue
                    setattr(sec_system, key, val)

            return sec_system

        def parse_calculation(source):
            if source is None:
                return
            sec_scc = parse_scc(source)
            sec_system = parse_system(source)
            sec_scc.single_configuration_calculation_to_system_ref = sec_system
            sec_scc.single_configuration_to_calculation_method_ref = sec_run.section_method[-1]

        # basis set correction
        # TODO determine if there is a need to add this
        basis_set_correction = calculation.get('basis_set_correction')
        if basis_set_correction is not None:
            for iteration in basis_set_correction.get('iteration', []):
                parse_calculation(iteration)

        # transition state search
        tss = calculation.get('tss')
        if tss is not None:
            for iteration in tss.get('iteration', []):
                parse_calculation(iteration)

        cg_refinement = calculation.get('cg_refinement')
        if cg_refinement is not None:
            for iteration in cg_refinement.get('iteration', []):
                parse_calculation(iteration)

        # molecular dyanamics
        # because of basis set correction number of frames may not correspond to number
        # of iterations
        md = calculation.get('md')
        if md is not None:
            # TODO add more quantities, read from md file
            for iteration in md.get('iteration', []):
                parse_calculation(iteration)

        # dmd optimization
        dmd = calculation.get('dmd')
        if dmd is not None:
            for iteration in dmd.get('iteration', []):
                parse_calculation(iteration)

        # di optimization
        di = calculation.get('di')
        if di is not None:
            for iteration in di.get('iteration', []):
                parse_calculation(iteration)

        # bfgs optimization
        bfgs = calculation.get('bfgs')
        if bfgs is not None:
            for iteration in bfgs.get('iteration', []):
                # an iteration can both have a pre-conditioning stage which we skip
                if iteration.get('iteration'):
                    parse_calculation(iteration.get('iteration')[-1])

        # final configuration
        if calculation.get('final') is not None:
            parse_calculation(calculation.get('final'))

        sampling_method = sec_run.section_sampling_method[0].sampling_method
        if sampling_method in ['single_point', 'phonon']:
            # single point calculation
            parse_calculation(calculation)

    def parse_parameters(self):
        section_map = {
            'basis set parameters': BasisSetCellDependent,
            'phonon parameters': x_castep_section_phonons,
            'electronic minimization parameters': x_castep_section_scf_parameters,
            'density mixing parameters': x_castep_section_density_mixing_parameters,
            'population analysis parameters': x_castep_section_population_analysis_parameters,
            'core level spectra parameters': x_castep_section_core_parameters,
            'band structure parameters': x_castep_section_band_parameters,
            'transition state search parameters': x_castep_section_ts_parameters,
            'optics parameters': x_castep_section_optics_parameters,
            'electronic spectroscopy parameters': x_castep_section_electronic_spectroscpy_parameters,
            'time-dependent dft parameters': x_castep_section_tddft_parameters}

        # TODO map all castep parameters
        sec_run = self.archive.section_run[0]
        title = self.out_parser.get('title', {})

        def create_section(definition):
            if definition is None:
                return

            create = True
            for section in sec_run.m_contents():
                if section.m_def == definition.m_def:
                    create = False
                    break
            return sec_run.m_create(definition) if create else section

        for key in title.keys():
            section = create_section(section_map.get(key))
            if section is None:
                continue
            for sub_key, val in title[key].items():
                sub_key = self._metainfo_map.get(key, {}).get(sub_key)
                if val is not None and sub_key is not None:
                    val = val.magnitude if hasattr(val, 'magnitude') else val
                    setattr(section, sub_key, val)

        # van der Waals
        dft_d = self.out_parser.get('dft_d')
        if dft_d is not None:
            sec_vdW = sec_run.m_create(x_castep_section_van_der_Waals_parameters)
            if dft_d.get('method') is not None:
                sec_vdW.x_castep_disp_method_name = dft_d.get('method')
            for parameter in dft_d.get('parameter', []):
                key = parameter[0].upper() if parameter[0] == 'lambda' else parameter[0]
                setattr(sec_vdW, 'x_castep_Parameter_%s' % key, float(parameter[1]))

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self.init_parser()

        sec_run = self.archive.m_create(Run)
        version = self.out_parser.get('program_version', ['CASTEP', ''])
        sec_run.program_name = version[0]
        if version[1]:
            sec_run.program_version = str(version[1])

        compilation = self.out_parser.get('program_compilation', '').strip().split(' ', 1)
        if len(compilation) == 2:
            sec_run.program_compilation_host = compilation[0]
            date, time = compilation[1].rsplit(' ', 1)
            sec_run.x_castep_program_compilation_date = date
            sec_run.x_castep_program_compilation_time = time

        for key in ['compiler', 'maths_library', 'fft_library', 'constants_reference']:
            val = self.out_parser.get(key)
            if val is not None:
                setattr(sec_run, 'x_castep_%s' % key, val)

        date_start = self.out_parser.get('run_start')
        if date_start is not None:
            date_start = datetime.strptime(date_start.strip(), '%d %b %Y %H:%M:%S')
            sec_run.time_run_date_start = (date_start - datetime(1970, 1, 1)).total_seconds()

        # basis set
        sec_run.program_basis_set_type = 'plane_waves'
        basis_parameters = self.out_parser.get('title', {}).get('basis set parameters', {})
        sec_basis_set = sec_run.m_create(BasisSetCellDependent)
        sec_basis_set.basis_set_cell_dependent_kind = 'plane_waves'
        cutoff = basis_parameters.get('plane wave basis set cut-off')
        if cutoff:
            sec_basis_set.basis_set_planewave_cutoff = cutoff
            sec_basis_set.basis_set_cell_dependent_name = 'PW_%d' % (round(cutoff.to('rydberg').magnitude))

        self.parse_method()

        self.parse_sampling_method()

        self.parse_topology()

        self.parse_configurations()

        self.parse_parameters()

        # times
        time = self.out_parser.get('time')
        if time is not None:
            sec_time = sec_run.m_create(x_castep_section_time)
            for key, val in time:
                setattr(sec_time, 'x_castep_%s_time' % key.lower(), val)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any, Tuple, Dict, Callable, Union, List
from io import BytesIO
import re

from nomad import utils

from .storage import ArchiveReader, ArchiveList, ArchiveObject


class ArchiveQueryError(Exception):
    '''
    An error that indicates that an archive query is either not valid or does not fit to
    the queried archive.
    '''
    pass


__query_archive_key_pattern = re.compile(r'^([\s\w\-]+)(\[([-?0-9]*)(:([-?0-9]*))?\])?$')


def query_archive(
        f_or_archive_reader: Union[str, ArchiveReader, BytesIO], query_dict: dict,
        **kwargs) -> Dict:
    '''
    Takes an open msg-pack based archive (either as str, reader, or BytesIO) and returns
    the archive as JSON serializable dictionary filtered based on the given required
    (query_dict) specification.

    Required example with some "required" (query_dict) features:

    .. code-block:: json
        {
            "results": "recursive-resolve-inplace",
            "workflow": {
                "final": "resolve"
            },
            "simulation": {
                "scc[-1]": {
                    "method": "recursive-resolve"
                    "system": {
                        "symmetry": "exclude",
                        "*": "include"
                    }
                    "dos": "recursive-include"
                }
            }
        }

    The different directives are:
        * include ('*' is an alias), includes whole subtree, resolves on a ref
        * recursive-resolve, includes whole subtree, resolves all refs recursively
        * recursive-resolve-inplace, includes whole subtree, resolves all refs recursively
            and replaces the ref with the resolved data
        * exclude (in combination with wildcard keys), replaces the value with null
    '''

    def _to_son(data):
        if isinstance(data, (ArchiveList, List)):
            data = [_to_son(item) for item in data]

        elif isinstance(data, ArchiveObject):
            data = data.to_dict()

        return data

    def _load_data(query_dict: Dict[str, Any], archive_item: ArchiveObject) -> Dict:
        query_dict_with_fixed_ids = {
            utils.adjust_uuid_size(key): value for key, value in query_dict.items()}
        return filter_archive(query_dict_with_fixed_ids, archive_item, transform=_to_son)

    if isinstance(f_or_archive_reader, ArchiveReader):
        return _load_data(query_dict, f_or_archive_reader)

    elif isinstance(f_or_archive_reader, (BytesIO, str)):
        with ArchiveReader(f_or_archive_reader, **kwargs) as archive:
            return _load_data(query_dict, archive)

    else:
        raise TypeError('%s is neither a file-like nor ArchiveReader' % f_or_archive_reader)


def filter_archive(
        required: Union[str, Dict[str, Any]], archive_item: Union[Dict, ArchiveObject, str],
        transform: Callable, result_root: Dict = None, resolve_inplace: bool = False) -> Dict:

    def _fix_index(index, length):
        if index is None:
            return index
        if index < 0:
            return max(-(length), index)
        else:
            return min(length, index)

    if archive_item is None:
        return None

    if isinstance(required, str):
        if required == 'exclude':
            return None

        if required == 'resolve':
            # TODO this requires to reflect on the definition to determine what are references!
            pass

        elif required in ['*', 'include']:
            pass

        else:
            raise ArchiveQueryError(f'unknown directive {required}')

        return transform(archive_item)

    elif not isinstance(required, dict):
        raise ArchiveQueryError('a value in required is neither dict not string directive')

    if isinstance(archive_item, str):
        # The archive item is a reference, the required is still a dict, the references
        # needs to be resolved
        # TODO
        raise ArchiveQueryError(f'resolving references in non partial archives is not yet implemented')

    result: Dict[str, Any] = {}
    for key, val in required.items():
        key = key.strip()

        # process array indices
        match = __query_archive_key_pattern.match(key)
        index: Union[Tuple[int, int], int] = None
        if match:
            key = match.group(1)

            # check if we have indices
            if match.group(2) is not None:
                first_index, last_index = None, None
                group = match.group(3)
                first_index = None if group == '' else int(group)

                if match.group(4) is not None:
                    group = match.group(5)
                    last_index = None if group == '' else int(group)
                    index = (0 if first_index is None else first_index, last_index)

                else:
                    index = first_index  # one item

            else:
                index = None

        elif key == '*':
            # TODO
            raise ArchiveQueryError('key wildcards not yet implemented')
        else:
            raise ArchiveQueryError('invalid key format: %s' % key)

        try:
            archive_child = archive_item[key]
            is_list = isinstance(archive_child, (ArchiveList, list))

            if index is None and is_list:
                index = (0, None)

            elif index is not None and not is_list:
                raise ArchiveQueryError('cannot use list key on none list %s' % key)

            if index is None:
                pass
            else:
                length = len(archive_child)
                if isinstance(index, tuple):
                    index = (_fix_index(index[0], length), _fix_index(index[1], length))
                    if index[0] == index[1]:
                        archive_child = [archive_child[index[0]]]
                    else:
                        archive_child = archive_child[index[0]: index[1]]
                else:
                    archive_child = [archive_child[_fix_index(index, length)]]

            if isinstance(archive_child, (ArchiveList, list)):
                result[key] = [
                    filter_archive(val, item, transform=transform)
                    for item in archive_child]
            else:
                result[key] = filter_archive(val, archive_child, transform=transform)

        except (KeyError, IndexError):
            continue

    return result



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from flask import Blueprint
from flask_restplus import Api


blueprint = Blueprint('api', __name__)

api = Api(
    blueprint,
    version='1.0', title='NOMAD API',
    description='Official NOMAD API',
    validate=True)
''' Provides the flask restplus api instance for the regular NOMAD api'''


# For some unknown reason it is necessary for each fr api to have a handler.
# Otherwise the global app error handler won't be called.
@api.errorhandler(Exception)
def errorhandler(error):
    '''When an internal server error is caused by an unexpected exception.'''
    return str(error)
